/* written by John Doug Reynolds, August 2001 */

#include "calib.h"
#include <CalServer.h>
#include <Convolver.h>
#include <errstr.h>
#include <algorithm>
#include <vector>

using namespace scanningLib;

class KernelSet {
public:
   KernelSet() : _padding(0) {}
   void load( const Spectrum &c, unsigned o );
   struct Spectrum& convolve( struct Spectrum& ) const;
private:
   struct Kernel {
      struct Spectrum curve;
      unsigned origin;
   };
   list<Kernel> _kernels;
   unsigned _padding;
   static vector<float>& wrap(vector<float>&,const Spectrum&,unsigned offset);
};

static void shift( vector<float>&, const Spectrum&, int offset );

static CcdQuantumEfficiencyCurve	ccdqeCurve[9];
static EffectiveAreaDataServer*		caldsEA;
static EffectiveAreaCurve*		eaCurve[6];
static GeometryDataServer*		caldsGM;
static GratingDataServer*		caldsGR;
static PsfDataServer*			caldsPS;
static GridSpan				betaGrid;
static double				theta, phi;

vector<double> vignetting;

/* --- LSF components ------------------------------------------------------ */

void prfBroad( vector<float> &rf,
	       float eLow, float eHigh, float alpha, int order ) 
{
   double energy = 1000 * (eLow + eHigh) / 2;
   double beta_m = caldsGR->beta(alpha,order,energy);
   int curveHead = betaGrid.cell(beta_m) - betaGrid.min;
   int right = max( (int) betaGrid.length() - curveHead - 1, 0 );
   int left = max( curveHead, 0 );

   Spectrum
      dist=caldsGR->scatteringDistribution(left, right, betaGrid.stride,
					   alpha, order, energy, false, true);
   shift( rf, dist, curveHead - left );

   float effective_area = eaCurve[-order]->area(1000*eLow,1000*eHigh);
   for ( unsigned i = 0; i < betaGrid.length(); ++i ) rf[i] *= effective_area;
}

void prfNarrow( vector<float> &rf,
		float eLow, float eHigh, float alpha, int order ) 
{
   float lambda = 2 * hc / (eLow+eHigh);
   float effective_area = eaCurve[-order]->area(1000*eLow,1000*eHigh);
   float beta_m = caldsGR->beta(alpha,order,1000*hc/lambda);
   float chrmgn = sin(alpha) / sin(beta_m);
   Spectrum dist;

   if ( betaGrid.contains( betaGrid.cell(beta_m) ) ) {
      struct KernelSet kernels;

      // small angle scattering

      dist = caldsGR->scatteringDistribution(3, betaGrid.stride, alpha, order,
					     1000*hc/lambda, true, false);
      DPair fSmallLarge = caldsGR->scatteringDistributionContribs();
      dist[dist.size()/2] += (1 - fSmallLarge.first)*(1 - fSmallLarge.second);
      kernels.load( dist, dist.size()/2 );

      // grating misalignment

      dist=caldsGR->misalignmentFigureDistribution(10,betaGrid.stride,chrmgn);
      kernels.load( dist, (unsigned)floor(.5+dist.firstMoment()) );

      // grating bow induced misalignment

      dist = caldsGR->bowingFigureDistribution( 10, betaGrid.stride, chrmgn );
      if ( dist.size() )
	 kernels.load( dist, (unsigned)floor(.5+dist.firstMoment()) );

      // geometric defocus approximation

      dist = caldsGM->lsfDefocusDistribution( 10, betaGrid.stride, theta, phi,
					      order, 1000*hc/lambda );
      kernels.load( dist, dist.size()/2 );

      // finite energy band contribution

      dist = caldsGR->broadeningDistribution(betaGrid.stride,
					     hc/eLow - hc/eHigh,beta_m,order);
      if ( dist.size() ) kernels.load( dist, dist.size()/2 );

      // mirror point spread contribution

      float centerIndex = betaGrid.tick(beta_m) - betaGrid.min + 0.5;
      float ico = centerIndex - floor(centerIndex) - 0.5;
      dist = caldsPS->dispersionFigureDistribution( 100, betaGrid.stride,
						    chrmgn, ico );
      int delta = (int)floor(centerIndex) - (int)floor(0.5+dist.firstMoment());

      shift( rf, kernels.convolve(dist), delta );
   }
   else {
      dist = caldsGR->scatteringDistribution(3,betaGrid.stride,alpha,
					     order,1000*hc/lambda,true,false);
      shift( rf, dist, betaGrid.cell(beta_m)-betaGrid.min-(int)dist.size()/2 );
   }

   for ( unsigned i = 0; i < betaGrid.length(); ++i ) rf[i] *= effective_area;
}

/* --- calibrations -------------------------------------------------------- */

float empiricalCrossLA( float energy, float dpsi, float xdsp0, float xdsp1 )
{
   float k = (energy * 10800) / (0.678 * M_PI);
   return (atan(k*(xdsp1+dpsi)) - atan(k*(xdsp0+dpsi))) / M_PI;
}

float calEnergy( float beta, int order, float alpha )
{
   return caldsGR->energy(alpha,beta,order);
}

float calAlpha0()
{
   return caldsGR->alpha0();
}

void ccd_qe( float qe[10], float energy /* keV */ )
{
   qe[0] = 0;
   for ( unsigned i = 1; i <= 9; ++i )
      qe[i] = ccdqeCurve[i-1].efficiency( energy * 1000 );
}

/* --- initialize/finalize ------------------------------------------------- */

void calibInit( const GridSpan &_betaGrid )
{
   betaGrid = _betaGrid;

   caldsGR = calServer.getAtom(caldsGR);
   caldsEA = calServer.getAtom(caldsEA);
   caldsPS = calServer.getAtom(caldsPS);
   caldsGM = calServer.getAtom(caldsGM);

   unsigned i;
   for ( i = 0; i < 6; ++i ) eaCurve[i] = 0;

   for ( i = 0; i < 9; ++i ) {
      struct CcdQuantumDataServer *qds=0;
      calServer.state()->ccd()->set(i+1);
      qds = calServer.getAtom(qds);
      ccdqeCurve[i] = qds->grade(1);
      calServer.releaseAtom(qds);
   }

   vector<double> beta(betaGrid.length());
   for ( i=0; i < betaGrid.length(); ++i ) beta[i] = betaGrid[i+betaGrid.min];
   vignetting = caldsGR->selfVignettingEfficiency(beta);
}

void resetSourcePosition( float _theta, float _phi )
{
   theta = _theta;
   phi = _phi;
   for ( unsigned i = 0; i < 6; ++i ) {
      delete eaCurve[i];
      eaCurve[i] = new EffectiveAreaCurve(
			caldsEA->intrinsicEffectiveAreaCurve(-i,theta,phi) );
   }
}

void calibFree()
{
   for ( unsigned i = 0; i < 6; ++i ) delete eaCurve[i];
   calServer.releaseAtom(caldsGM);
   calServer.releaseAtom(caldsPS);
   calServer.releaseAtom(caldsEA);
   calServer.releaseAtom(caldsGR);
}

/* --- utility ------------------------------------------------------------- */

void shift( vector<float> &curve, const Spectrum &dist, int offset )
{
   float *ptr = curve.begin();

   int begin = max( offset, 0 );
   int end = min( (int)dist.size() + offset, (int)curve.size() );

   if ( end < begin ) {
      memset( ptr, 0, curve.size() * sizeof(float) );
      return;
   }

   if ( begin > 0 ) {
      memset( ptr, 0, begin * sizeof(float) );
      ptr += begin;
   }

   vector<double>::const_iterator itr, endI = dist.begin() + end - offset;
   for ( itr = dist.begin() + begin - offset; itr != endI; *ptr++ = *itr++ );

   if ( (int)curve.size() > end )
      memset( ptr, 0, (curve.size() - end) * sizeof(float) );
}

/* --- KernelSet methods --------------------------------------------------- */

void KernelSet::load( const Spectrum &c, unsigned o )
{
   _kernels.push_back( Kernel() );
   _kernels.back().curve = c;
   _kernels.back().origin = o;
   maximize( _padding, o );
   maximize( _padding, (unsigned)c.size() - o );
}

struct Spectrum& KernelSet::convolve( struct Spectrum &dist ) const
{
   struct Convolver
      convolver( 1 << (int) ceil( log(dist.size()+_padding) / log(2) ) );

   vector<float> curve( convolver.size() );
   shift( curve, dist, 0 );
   convolver.load( curve );

   list<Kernel>::const_iterator itr;
   for ( itr = _kernels.begin(); itr != _kernels.end(); ++itr )
      convolver.load( wrap(curve,itr->curve,itr->origin) );

   dist.assign( convolver().begin(), convolver().begin() + dist.size() );
   return dist;
}

vector<float>&
KernelSet::wrap( vector<float> &curve, const Spectrum &dist, unsigned offset )
{
   vector<double>::const_iterator center, begin, end, itr;

   if ( dist.size() > curve.size() ) {
      begin = dist.begin() + (dist.size() - curve.size())/2;
      end = begin + curve.size();
   }
   else {
      begin = dist.begin();
      end = dist.end();
   }

   center = dist.begin() + offset;
   if ( center < begin )
      center = begin;
   else if ( center > end )
      center = end;

   float *ptr = curve.begin();
   for ( itr = center; itr != end; *ptr++ = *itr++ );

   unsigned count = curve.size() - (end-begin);
   memset( ptr, 0, count * sizeof(float) );
   ptr += count;

   for ( itr = begin; itr != center; *ptr++ = *itr++ );

   return curve;
}
