/* written by John Doug Reynolds, June 2001 */

#include "rgsrmfgen.h"
#include "calib.h"
#include <ShapeScanner.h>
#include <rgsPixelsLib.h>
#include <DalVec.h>
#include <ASCregion.h>
#include <CalServer.h>
#include <cstdio>

using namespace rgsPixelsLib;

struct LossStats {
   list<Span>			spans;		// extraction column segments
   float			exposure;	// normalized exposure
   float			crossPSF;	// normalized exposure*crossPSF
   void ingest( const ExposureMap&, const Spectrum&, int column, int offset );
   void normalize( float norm ) { exposure *= norm; crossPSF *= norm; }
   LossStats() : exposure(0), crossPSF(0) {}
};

struct LossFunction {
   unsigned char		ccd;		// CCD identifier
   unsigned			offset;		// beta offset of array origin
   vector<LossStats>		fg, bg;		// indexed by beta subspace
};

class OutRmfOGIP {
public:
   OutRmfOGIP( struct DataSet *dataset, const vector<float> &ebounds,
	       const list<const Attribute*>&, const string &spectrum,
	       float threshold, bool corrected, unsigned size
	       );
   void operator()( const vector<float>&, float eLo, float eHi );
   void release();
private:
   DalVecI _nGrp;
   DalVecE _energLo, _energHi;
   struct Column *_fChan, *_nChan, *_matrix;
   unsigned _size, _next;
   float _threshold;
};

void OutRmfOGIP::release()
{
   _nGrp.release();
   _energLo.release();
   _energHi.release();
}

void OutRmfOGIP::operator()( const vector<float> &lsf, float eLo, float eHi )
{
   if ( _next >= _size )
      Sas::error("matrixBounds","attempt to add bin to completed matrix");

   vector< pair<unsigned,unsigned> > group;
   unsigned i, nval = 0;
   bool outside = true;

   for ( i = 0; i < lsf.size(); ++i )
      if ( lsf[i] >= _threshold ) {
	 if ( outside ) {
	    group.push_back( pair<unsigned,unsigned>(i,0) );
	    outside = false;
	 }
	 ++group.back().second;
	 ++nval;
      }
      else outside = true;

   _nGrp[_next] = group.size();
   _energLo[_next] = eLo;
   _energHi[_next] = eHi;

   struct CellData *fChanData = _fChan->cellData(_next);
   struct CellData *nChanData = _nChan->cellData(_next);
   struct CellData *matrixData= _matrix->cellData(_next);

   fChanData->size(group.size());
   nChanData->size(group.size());
   matrixData->size(nval);
      
   float total = 0, *dst = matrixData->real32Data();
   for ( i = 0; i < group.size(); ++i ) {
      (fChanData->int16Data())[i] = group[i].first + 1;
      (nChanData->int16Data())[i] = group[i].second;
      vector<float>::const_iterator src = lsf.begin() + group[i].first;
      for ( unsigned count = group[i].second; count--; *dst++ = *src++ )
	 total += *src;
   }

   delete matrixData;
   delete nChanData;
   delete fChanData;
   ++_next;

   errstr << "row " << _next << " (" << (eLo+eHi)/2 << " keV): "
	  << "total response " << total << " in " << group.size()
	  << " group" << (group.size() == 1 ? "" : "s")
	  << Sas::message(Msg::App,Msg::Sparse);

   if ( group.size() ) {
      errstr << "    groups:";
      for ( i = 0; i < group.size(); ++i )
	 errstr << ' ' << group[i].first + 1 << '-'
		<< group[i].first + group[i].second;
      errstr << Sas::message(Msg::App,Msg::Verbose);
   }
}

OutRmfOGIP::OutRmfOGIP( struct DataSet *dataset,
			const vector<float> &ebounds,
			const list<const Attribute*> &attributes,
			const string &spectrum,
			float threshold,
			bool corrected,
			unsigned size
			)
   : _size(size), _next(0), _threshold(threshold)
{
   unsigned detchans = ebounds.size() - 1;

   struct Table *table = dataset->addTable( "MATRIX", _size );

   _energLo.reset( table->addColumn("ENERG_LO",Dal::Real32,"","keV") );
   _energHi.reset( table->addColumn("ENERG_HI",Dal::Real32,"","keV") );
   _nGrp.reset( table->addColumn("N_GRP",Dal::Int16,"","") );

   _fChan = table->addColumn("F_CHAN",Dal::Int16,"","",Dal::zero);
   _nChan = table->addColumn("N_CHAN",Dal::Int16,"","",Dal::zero);
   _matrix = table->addColumn("MATRIX",Dal::Real32,"","",Dal::zero);

   *_fChan->addAttribute( "TLMIN" ) = 1;
   *_fChan->addAttribute( "TLMAX" ) = int(detchans);

   list<const Attribute*>::const_iterator iter;
   for ( iter = attributes.begin(); iter != attributes.end(); ++iter )
      table->addAttribute( *iter );

   *table->addAttribute( "FILTER" ) = "NONE";
   *table->addAttribute( "CHANTYPE" ) = "PI";
   *table->addAttribute( "DETCHANS" ) = int(detchans);
   *table->addAttribute( "HDUCLASS" ) = "OGIP";
   *table->addAttribute( "HDUCLAS1" ) = "RESPONSE";
   *table->addAttribute( "HDUCLAS2",
			 corrected ? "NET response" : "TOTAL response"
			 ) = "RSP_MATRIX";
   *table->addAttribute( "HDUCLAS3", "all effects included" ) = "FULL";
   *table->addAttribute( "HDUVERS" ) = "1.3.0";
   *table->addAttribute( "PHAFILE", "corresponding spectrum file" ) = spectrum;
   *table->addAttribute( "LO_THRES", "threshold of non-zero response"
			 ) = _threshold;
			 
   *table->addAttribute( "RMFVERSN", "Obsolete" ) = "1992a";
   *table->addAttribute( "HDUVERS1", "Obsolete" ) = "1.1.0";
   *table->addAttribute( "HDUVERS2", "Obsolete" ) = "1.2.0";

   // add the EBOUNDS extension

   table = dataset->addTable( "EBOUNDS", detchans );

   DalVecI channel(table->addColumn("CHANNEL",Dal::Int16,"","channel"));
   DalVecE eMin(table->addColumn("E_MIN",Dal::Real32,"","keV"));
   DalVecE eMax(table->addColumn("E_MAX",Dal::Real32,"","keV"));

   for ( iter = attributes.begin(); iter != attributes.end(); ++iter )
      table->addAttribute( *iter );

   *table->addAttribute( "FILTER" ) = "NONE";
   *table->addAttribute( "CHANTYPE" ) = "PI";
   *table->addAttribute( "DETCHANS" ) = int(detchans);
   *table->addAttribute( "HDUCLASS" ) = "OGIP";
   *table->addAttribute( "HDUCLAS1" ) = "RESPONSE";
   *table->addAttribute( "HDUCLAS2" ) = "EBOUNDS";
   *table->addAttribute( "HDUVERS" ) = "1.2.0";
   *table->addAttribute( "PHAFILE", "corresponding spectrum file" ) = spectrum;

   *table->addAttribute( "RMFVERSN", "Obsolete" ) = "1992a";
   *table->addAttribute( "HDUVERS1", "Obsolete" ) = "1.0.0";
   *table->addAttribute( "HDUVERS2", "Obsolete" ) = "1.1.0";

   for ( unsigned i = 0; i < detchans; ++i ) {
      channel[i] = i+1;
      eMin[i] = ebounds[i+1];
      eMax[i] = ebounds[i];
   }
}

struct ShapeScanner* scanner( struct Table*,
			      const Grid &xGrid, const Grid &yGrid );

void init( const Table*, int sourceid, float &alpha, float &dpsi );

void init( list<LossFunction>&, struct DataSet*, struct ShapeScanner &fg,
	   struct ShapeScanner *bg, float dpsi );

void init( vector<Span> &banana, struct ShapeScanner& );

float empiricalCrossLA( float energy, float dpsi, const list<Span>& );

void nodeEfficiencySA( vector<float>&, float energy );
void nodeEfficiencyLA( vector<float>&, float dpsi, float energy );
void betaEfficiency( vector<float>&, float energy );

bool subdivides( const GridSpan &fine, const GridSpan &coarse );

vector<Span>		banana;
list<LossFunction>	lossByNode;
struct GridSpan		betaGrid, xdspGrid;

void rgsrmfgen()
{
   string filename = stringParameter("file");
   string dataname = stringParameter("set");
   string sourcename = stringParameter("sourceset");
   unsigned source = (unsigned) intParameter("source");
   int order = intParameter("order");
   unsigned ebins = (unsigned) intParameter("ebins");
   float emin = realParameter("emin");
   float emax = realParameter("emax");
   bool withbackground = booleanParameter("withbackground");
   bool withspectrum = booleanParameter("withspectrum");
   string spectrumname = stringParameter("spectrumset");

   string basename;
   if ( dataname.size() > 25 ) basename = dataname.substr(0,25);

   struct DataSet *dataset = dataSetServer->open( dataname, Dal::Read );
   struct Table *etab = dataset->table("EVENTS");

   if ( source == 0 ) source = etab->attribute("SOURCEID")->asInt();

   // initialize the coordinate grids

   betaGrid = gridspan( etab->column("BETA_CHANNEL") );
   xdspGrid = gridspan( etab->column("XDSP_CHANNEL") );

   if ( withspectrum ) {
      if ( spectrumname == "" )
	 if ( basename.size() ) {
	    char extension[32];
	    sprintf( extension, "%s%do%d.pha",
		     withbackground ? "tot" : "net", source, order );
	    spectrumname = basename + extension;
	 }
	 else
	    Sas::error("noSpectrumFile",
		       "a default spectrumset cannot be inferred from set: "
		       "specify the filename explicitly");

      struct DataSet *specset = dataSetServer->open( spectrumname, Dal::Read );
      const Table    *spectab = specset->table("SPECTRUM");

      string hduclas2 = spectab->attribute("HDUCLAS2")->asString();
      if ( hduclas2 == "BKG" )
	 Sas::error("bkgResponse",
		    "cannot make a response for this kind of spectrum (BKG)");

      withbackground = (hduclas2 != "NET");
      source = spectab->attribute("SOURCEID")->asInt();
      order = spectab->attribute("RFLORDER")->asInt();

      struct GridSpan chnGrid = gridspan( spectab->column("CHANNEL") );
      if ( !subdivides(betaGrid,chnGrid) )
	 Sas::error("badRebinning",
		    "SPECTRUM beta channels not compatible with EVENTS");
      betaGrid = chnGrid;

      dataSetServer->close( specset );
   }

   // other initializations

   if ( filename == "" )
      if ( basename.size() ) {
	 char extension[32];
	 sprintf( extension, "%s%do%d.rmf",
		  withbackground ? "tot" : "net", source, order );
	 filename = basename + extension;
      }
      else
	 Sas::error("noResultFile",
		    "a default file cannot be inferred from set: "
		    "specify the filename explicitly");

   if ( sourcename == "" )
      if ( basename.size() )
	 sourcename = basename + "sources.ds";
      else
	 Sas::error("noSourcesFile",
		    "a default sourceset cannot be inferred from set: "
		    "specify the filename explicitly");

   string instrument = dataset->attribute("INSTRUME")->asString();
   calServer.setInstrument( Xmm::toInstrument(instrument) );

   const float low_threshold = 1e-7; // smallest non-zero response
   calibInit( betaGrid );

   struct DataSet *sourceset = dataSetServer->open( sourcename, Dal::Read );

   float alpha, dpsi;
   init( sourceset->table("SRCLIST"), source, alpha, dpsi );

   char extname[19];
   struct ShapeScanner *foreground, *background, *pulseheight;

   sprintf( extname, "%s_SRC%d_SPATIAL", instrument.c_str(), source );
   foreground = scanner( sourceset->table(extname), betaGrid, xdspGrid );

   if ( !withbackground ) {
      sprintf( extname, "%s_BACKGROUND", instrument.c_str() );
      background = scanner( sourceset->table(extname), betaGrid, xdspGrid );
   }
   else background = 0;

   sprintf( extname, "%s_SRC%d_ORDER_%d", instrument.c_str(), source, order );
   pulseheight = scanner( sourceset->table(extname), betaGrid, Grid() );

   init( lossByNode, dataset, *foreground, background, dpsi );
   init( banana, *pulseheight );

   delete pulseheight;
   delete background;
   delete foreground;

   dataSetServer->close( sourceset );

   // construct the output dataset

   struct DataSet *rmfset = dataSetServer->open( filename, Dal::Create );

   *rmfset->addAttribute( "SOURCEID",
			  "SRCLIST:INDEX source identifier" ) = int(source);
   *rmfset->addAttribute( "RFLORDER", "ordinal reflection order" ) = order;

   list<const Attribute*> attributes;
   attributes.push_back( dataset->attribute("TELESCOP") );
   attributes.push_back( dataset->attribute("INSTRUME") );

   vector<float> ebounds( betaGrid.length() + 1 );
   for ( int i = betaGrid.min; i <= betaGrid.max+1; ++i )
      ebounds[i-betaGrid.min] = .001*calEnergy(betaGrid[i-.5],-order,alpha);

   struct OutRmfOGIP rmf( rmfset, ebounds, attributes,
			  withspectrum ? spectrumname : "NONE",
			  low_threshold, !withbackground, ebins );

   vector<float> lsf( betaGrid.length() );
   vector<float> accumulator( betaGrid.length() );
   vector<float> response( betaGrid.length() );

   float lambda = hc/emin, deltaLambda = (hc/emin - hc/emax) / ebins;

   for ( unsigned j = 0; j < ebins; ++j ) {
      size_t i, bgSize = betaGrid.length();
      float eLow, eHigh;

      eLow = hc/lambda; lambda -= deltaLambda; eHigh = hc/lambda;

      float energy = (eLow + eHigh) / 2;

      // small angles

      for ( i = 0; i < bgSize; ++i ) accumulator[i] = 0;
      for ( int ord = 0; ord < 6; ++ord ) {
	 prfNarrow( lsf, eLow, eHigh, alpha, -ord );
	 for ( i = 0; i < bgSize; ++i ) accumulator[i] += lsf[i];
      }
      nodeEfficiencySA( accumulator, energy );
      for ( i = 0; i < bgSize; ++i ) response[i] = accumulator[i];

      // large angles

      for ( i = 0; i < bgSize; ++i ) accumulator[i] = 0;
      for ( int ord = 0; ord < 6; ++ord ) {
	 prfBroad( lsf, eLow, eHigh, alpha, -ord );
	 for ( i = 0; i < bgSize; ++i ) accumulator[i] += lsf[i];
      }
      nodeEfficiencyLA( accumulator, dpsi, energy );
      for ( i = 0; i < bgSize; ++i ) response[i] += accumulator[i];

      //

      betaEfficiency( response, energy );
      rmf( response, eLow, eHigh );
   }

   rmf.release();

   dataSetServer->close( rmfset );
   dataSetServer->close( dataset );

   calibFree();
}

struct ShapeScanner* scanner( struct Table *table,
			      const Grid &xGrid, const Grid &yGrid )
{
   struct GeoRegion geoReg;
   struct InASCregionTable ascReg(table);
   ascReg >> geoReg;
   return new RegionScanner( geoReg, xGrid, yGrid );
}

void init( const Table *table, int sourceid, float &alpha, float &dpsi )
{
   const DalVecI index( table->column("INDEX") );

   for ( unsigned i = 0; i < table->rows(); ++i )
      if ( index[i] == sourceid ) {
	 resetSourcePosition( DalVecE(table->column("FOV_R"))[i]*3600,
			      DalVecE(table->column("FOV_PHI"))[i]*M_PI/180 );

	 alpha=calAlpha0()+DalVecE(table->column("DELTA_DISP"))[i]*M_PI/10800;
	 dpsi = DalVecE(table->column("DELTA_XDSP"))[i]*M_PI/10800;

	 return;
      }

   Sas::error("badSource","source is not present in SRCLIST table");
}

void LossStats::ingest( const ExposureMap &emap, const Spectrum &xpsf,
			int column, int offset )
{
   list<Span>::const_iterator spanItr;
   for ( spanItr = spans.begin(); spanItr != spans.end(); ++spanItr ) {
      struct Span span = Span(offset+spanItr->min,offset+spanItr->max);
      if ( !(span &= emap.bounds().row) ) continue;
      long itr = emap.id(column,span.min-1), end = emap.id(column,span.max);
      for ( int i = span.min - offset; itr != end; ) {
	 float expsr = emap[ emap.incr(itr,Compass::North) ];
	 crossPSF += xpsf[i++] * expsr;
	 exposure += expsr;
      }
   }
}

void init( list<LossFunction> &lbn, struct DataSet *dataset,
	   struct ShapeScanner &fgScanner, struct ShapeScanner *bgScanner,
	   float dpsi )
{
   struct CanonicalCrossPsf *crossPSF = calServer.getAtom(crossPSF);

   CalReal64Vector xdspBins( 0, xdspGrid.length() );
   for ( int i = xdspGrid.min; i <= xdspGrid.max; ++i )
      xdspBins[i-xdspGrid.min] = xdspGrid[i];

   //

   unsigned validCount =  0;
   for ( unsigned ccd = 1; ccd <= 9; ++ccd )
      for ( unsigned node = 0; node < 2; ++node ) {
	 char extname[9];
	 sprintf( extname, "EXPMAP%u%u", node, ccd );
	 if ( !dataset->hasBlock(extname) ) continue;
	 struct ExposureMap emap( dataset->array(extname) );

	 bool valid = true;
	 if ( !subdivides(GridSpan(emap.beta,emap.bounds().col),betaGrid) ) {
	    errstr << "the exposure map dispersion channels are not "
		   << "compatible with the EBOUNDS table";
	    valid = false;
	 }
	 if ( !commensurate(emap.xdsp,xdspGrid) ) {
	    errstr << "the exposure map cross-dispersion channels are not "
		   << "commensurate with the EVENTS table";
	    valid = false;
	 }
	 if ( !emap.bounds() ) {
	    errstr << "the exposure map is null";
	    valid = false;
	 }

	 if ( !valid ) {
	    errstr << ": discarding " << extname << Sas::warning("invalidMap");
	    continue;
	 }
	 else ++validCount;

	 // initialize the clipping region

	 struct Bounds clip;
	 clip.col = Span( betaGrid.cell( emap.beta[emap.bounds().col.min] ),
			  betaGrid.cell( emap.beta[emap.bounds().col.max] ) );
	 clip.row = Span( xdspGrid.cell( emap.xdsp[emap.bounds().row.min] ),
			  xdspGrid.cell( emap.xdsp[emap.bounds().row.max] ) );
	 clip &= Bounds(betaGrid,xdspGrid);

	 // initialize the loss function

	 size_t size = clip.col.length();

	 lbn.push_back( LossFunction() );
	 struct LossFunction &loss = lbn.back();
	 loss.ccd = ccd;
	 loss.offset = clip.col.min - betaGrid.min;

	 // initialize the accumulators

	 list<Span> spans;
	 int b;

	 loss.fg.resize(size);
	 fgScanner.reset(&clip);
	 while ( fgScanner.next(b,spans) )
	    loss.fg[b-clip.col.min].spans = spans;

	 if ( bgScanner ) {
	    loss.bg.resize(size);
	    bgScanner->reset(&clip);
	    while ( bgScanner->next(b,spans) )
	       loss.bg[b-clip.col.min].spans = spans;
	 }

	 // accumulate over this node

	 const Bounds &bnds = emap.bounds();
	 int offset = bnds.row.min - xdspGrid.cell( emap.xdsp[bnds.row.min] );
	 int prev = clip.col.min - 1;
	 struct Spectrum dist;

	 for ( int column = bnds.col.min; column <= bnds.col.max; ++column ) {
	    b = betaGrid.cell( emap.beta[column] );
	    if ( b < clip.col.min ) continue;
	    else if ( b > clip.col.max ) break;

	    if ( b != prev ) {
	       dist = crossPSF->probability(0,betaGrid[b],dpsi,
					    xdspBins,xdspGrid.stride);
	       prev = b;
#if 0
	       {
		  static bool once = false;
		  if ( !once && b == 1700 ) {
		     struct Span &span = loss.fg[b-clip.col.min].spans.front();
		     printf("! beta=%e dpsi=%e xdsp0=%e xdsp1=%e\n",
			    betaGrid[b], dpsi,
			    xdspGrid[span.min], xdspGrid[span.max]);
		     for ( unsigned i = 0; i < dist.size(); ++i )
			printf("%15e%15e%15e\n",xdspBins[i],dist[i],
			       empiricalCrossLA(1.525,dpsi,
						xdspBins[0],xdspBins[i]));
		     once = true;
		  }
	       }
#endif
	    }

	    loss.fg[b-clip.col.min].ingest( emap, dist, column, offset );
	    if ( bgScanner )
	       loss.bg[b-clip.col.min].ingest( emap, dist, column, offset );
	 }

	 // normalize the stats

	 float datamax = emap.array->attribute("DATAMAX")->asReal();
	 float normalizer = emap.beta.stride / betaGrid.stride / datamax;

	 vector<LossStats>::iterator itr;
	 for ( itr = loss.fg.begin(); itr != loss.fg.end(); ++itr )
	    itr->normalize( normalizer );

	 if ( bgScanner )
	    for ( itr = loss.bg.begin(); itr != loss.bg.end(); ++itr )
	       itr->normalize( normalizer );
      }

   if ( validCount == 0 )
      Sas::error("nullExposure","no valid exposure maps are present");

   calServer.releaseAtom(crossPSF);
}

void init( vector<Span> &banana, struct ShapeScanner &scanner )
{
   int b;
   list<Span> spans;
   struct Bounds clip( betaGrid, Span(0,scanner.bounds().row.max) );

   scanner.reset( &clip );
   banana.resize( betaGrid.length() );

   while ( scanner.next(b,spans) )
      if ( spans.size() > 1 )
	 Sas::error("badBanana","banana has multiple spans per beta");
      else
	 banana[b-betaGrid.min] = spans.front();
}

void nodeEfficiencySA( vector<float> &rf, float energy )
{
   vector<float> lossf( betaGrid.length() );

   float qe[10];
   ccd_qe( qe, energy );

   list<LossFunction>::const_iterator itr;
   for ( itr = lossByNode.begin(); itr != lossByNode.end(); ++itr ) {
      size_t b = itr->offset;
      const float ccdqe = qe[ itr->ccd ];
      vector<LossStats>::const_iterator fgI, bgI = itr->bg.begin();
      for ( fgI = itr->fg.begin(); fgI != itr->fg.end(); ++fgI ) {
	 if ( b >= lossf.size() ) break; // necessary?
	 float xpsf = fgI->crossPSF;
	 if ( bgI != itr->bg.end() ) {
	    float scale = bgI->exposure ? fgI->exposure / bgI->exposure : 0;
	    xpsf -= scale * (bgI++)->crossPSF;
	 }
	 lossf[b++] += xpsf * ccdqe;
      }
   }

   for ( size_t i = 0; i < rf.size(); ++i ) rf[i] *= lossf[i];
}

void nodeEfficiencyLA( vector<float> &rf, float dpsi, float energy )
{
#if 0
   printf("! energy=%e dpsi=%e\n",energy,dpsi);
#endif
   vector<float> lossf( betaGrid.length() );

   float qe[10];
   ccd_qe( qe, energy );

   list<LossFunction>::const_iterator itr;
   for ( itr = lossByNode.begin(); itr != lossByNode.end(); ++itr ) {
      size_t b = itr->offset;
      const float ccdqe = qe[ itr->ccd ];
      vector<LossStats>::const_iterator fgI, bgI = itr->bg.begin();
      for ( fgI = itr->fg.begin(); fgI != itr->fg.end(); ++fgI ) {
	 if ( b >= lossf.size() ) break; // necessary?

	 float xpsf = empiricalCrossLA( energy, dpsi, fgI->spans );
	 if ( bgI != itr->bg.end() )
	    xpsf -= empiricalCrossLA( energy, dpsi, (bgI++)->spans );

#if 0
	 if ( fgI->spans.size() )
	    printf("%15e%15e%15e%15e\n",betaGrid[b],
		   xdspGrid[fgI->spans.back().min],
		   xdspGrid[fgI->spans.back().max],
		   xpsf
		   );
#endif
	 lossf[b++] += fgI->exposure * xpsf * ccdqe;
      }
   }

   for ( size_t i = 0; i < rf.size(); ++i ) rf[i] *= lossf[i];
}

void betaEfficiency( vector<float> &rf, float energy )
{
   const float energy_ph_tolerance = 0.01; // keV

   static float energy0 = -hc;
   static vector<float> lossf( betaGrid.length() );
   size_t i;

   if ( fabs(energy0 - energy) > energy_ph_tolerance ) {
      energy0 = energy + (energy > energy0 ? 1 : -1) * energy_ph_tolerance;

      calServer.state()->ccd()->set(1); // any CCD is fine
      struct CanonicalRedist *redist = calServer.getAtom(redist);
      struct Spectrum dist( redist->spectrum(1000*energy0) );
      calServer.releaseAtom(redist);

      float sum = 0;
      vector<float> cdist( dist.size() );
      for ( i = 0; i < dist.size(); ++i ) cdist[i] = (sum += dist[i]);

      for ( i = 0; i < lossf.size(); ++i ) {
	 int eMin = banana[i].min-1, eMax = banana[i].max;
	 if ( eMax >= (int) cdist.size() ) eMax = (int) cdist.size() - 1;
	 lossf[i] = eMin<eMax ? cdist[eMax] - (eMin<0 ? 0 : cdist[eMin]) : 0;
      }
   }

   for ( i = 0; i < rf.size(); ++i ) rf[i] *= vignetting[i] * lossf[i];
}

float empiricalCrossLA( float energy, float dpsi, const list<Span> &spans )
{
   int n = 0;
   float s = 0;
   list<Span>::const_iterator itr;
   for ( itr = spans.begin(); itr != spans.end(); ++itr ) {
      s += empiricalCrossLA( energy, dpsi,
			     xdspGrid[itr->min-.5], xdspGrid[itr->max+.5] );
      n += itr->length();
   }
   return s/n;
}

bool multiple( float multiple, float divisor, float scale )
{
   float count = floor( 0.5 + multiple / divisor );
   return fabs(multiple - count * divisor) <= 5e-7 * fabs(scale);
}

bool subdivides( const GridSpan &fine, const GridSpan &coarse )
{
   return
      multiple( coarse.stride, fine.stride, coarse.stride )
      &&
      multiple( fine.minval() - coarse.minval(), fine.stride,
		max(fabs(fine.origin),fabs(coarse.origin)) );
}
